//*******************************  Библиотеки  ******************************//
#include "Arduino.h"

#include "Config.h"

#include "UART.h"

#include "Type.h"

#include "FastIO.h"

#include "Protocol.h"
//*******************************  /Библиотеки  *****************************//

//********************  Объявление переменных, констант  ********************//
// Структура хранящая текущие значения нажатых кнопок и положения стиков джойстика от PS2.
ps2Struct ps2S;

// Структура хранящая предыдущие значения нажатых кнопок и положения стиков джойстика от PS2.
rovDataStruct rovDataS;

static uint8_t CurrentUARTPortSend = UART_PORT;
static uint8_t CurrentUARTPortRead = 0;

// Буфер хранящий сами данные без служебной информации.
#define InBufferSize 64
static uint8_t InBuffer[InBufferSize][UARTNumber];
static uint8_t IndexBuffer[UARTNumber];

// Переменные служебной информации.
static uint8_t CmdDataSize[UARTNumber];
static uint8_t CmdCheckSum[UARTNumber];
static uint8_t Cmd[UARTNumber];
//********************  /Объявление переменных, констант  *******************//

//**********************  Объявление приватные функции  *********************//
/// <summary>
/// Сохраняем отправляемый 8 бит в буфер TX и считаем контрольную сумму.
/// </summary>
static void Serialize8(uint8_t a) 
{
  UARTSerialize(CurrentUARTPortSend,a);
  CmdCheckSum[CurrentUARTPortSend] ^= a;
}

/// <summary>
/// Сохраняем отправляемый 16 бит в буфер TX и считаем контрольную сумму.
/// </summary>
static void Serialize16(uint16_t a) 
{
  Serialize8((a   ) & 0xFF);
  Serialize8((a>>8) & 0xFF);
}

/// <summary>
/// Сохраняем отправляемый 32 бит в буфер TX и считаем контрольную сумму.
/// </summary>
static void Serialize32(uint32_t a) 
{
  Serialize8((a    ) & 0xFF);
  Serialize8((a>> 8) & 0xFF);
  Serialize8((a>>16) & 0xFF);
  Serialize8((a>>24) & 0xFF);
}

/// <summary>
/// Читаем 8 бит данных из буфера.
/// </summary>
static uint8_t Read8()  
{
  return InBuffer[IndexBuffer[CurrentUARTPortRead]++][CurrentUARTPortRead]&0xff;
}

/// <summary>
/// Читаем 16 бит данных из буфера.
/// </summary>
static uint16_t Read16() 
{
  uint16_t t = Read8();
  t+= (uint16_t)Read8()<<8;
  return t;
}

/// <summary>
/// Читаем 32 бит данных из буфера.
/// </summary>
static uint32_t Read32() 
{
  uint32_t t = Read16();
  t+= (uint32_t)Read16()<<16;
  return t;
}

/// <summary>
/// Читаем структуру из буфера.
/// </summary>
static void ReadStruct(uint8_t *cb,uint8_t siz) 
{
  while(siz--) 
  {
    *cb++ = Read8();
  }
}

/// <summary>
/// Конвертируем байты в структуру.
/// </summary>
static void ConversionBytesIntoStructure() 
{  
  switch(Cmd[CurrentUARTPortRead])
  {
    case 1:
      ReadStruct((uint8_t*)&ps2S,12);
      break;
    case 2:
      ReadStruct((uint8_t*)&rovDataS,1);
      break;
  }    
}
//**********************  /Объявление приватные функции *********************//

//**********************  Объявление публичных функции  *********************//
/// <summary>
/// Сохраняем структуру в буфер TX и считаем контрольную сумму.
/// </summary>
void SerializeStruct(uint8_t *cb,uint8_t siz) {
  while(siz--) 
  {
    Serialize8(*cb++);
  }
}

/// <summary>
/// Сохраняем заголовок команды в буфер TX и считаем контрольную сумму.
/// </summary>
void HeadCmdSend(uint8_t err, uint8_t s, uint8_t c) {
  Serialize8('?');
  Serialize8('C');
  
  CmdCheckSum[CurrentUARTPortSend] = 0;
  
  Serialize8(s);
  Serialize8(c);
}

/// <summary>
/// Сохраняем хвост команды в буфер TX, считаем контрольную сумму и отправляем данные.
/// </summary>
void TailCmdSend() 
{
  Serialize8(CmdCheckSum[CurrentUARTPortSend]);
  UARTSendData(CurrentUARTPortSend);
}

/// <summary>
/// Получение данных.
/// </summary>
void UARTGetData()
{
  uint8_t c, n;  
  static uint8_t InBufferOffset[UARTNumber];
  static enum _CmdState 
  {
    IDLE,
    HEADER_START,
    HEADER_SIZE,
    HEADER_CMD,
    HEADER_DATA
  } CmdState[UARTNumber];
  
  for(n=0; n < UARTNumber; n++) 
  {
    #if !defined(ARDUINO_PRO_MINI)
      CurrentUARTPortRead = n;
    #endif
    
    uint8_t cc = UARTAvailable(CurrentUARTPortRead);
    
    while (cc--) 
    {
      c = UARTRead(CurrentUARTPortRead);
      
      if (CmdState[CurrentUARTPortRead] == IDLE) 
      {
            CmdState[CurrentUARTPortRead] = (c=='?') ? HEADER_START : IDLE;
            
      }
      else if (CmdState[CurrentUARTPortRead] == HEADER_START) 
      { 
            CmdState[CurrentUARTPortRead] = (c=='C') ? HEADER_SIZE : IDLE;
      }
      else if (CmdState[CurrentUARTPortRead] == HEADER_SIZE) 
      {
            if (c > InBufferSize) 
            {  
              CmdState[CurrentUARTPortRead] = IDLE;
              continue;
            }
            CmdDataSize[CurrentUARTPortRead] = c;
            
            CmdCheckSum[CurrentUARTPortRead] = 0;
            IndexBuffer[CurrentUARTPortRead] = 0;
            InBufferOffset[CurrentUARTPortRead] = 0;
            
            CmdCheckSum[CurrentUARTPortRead] ^= c;
            CmdState[CurrentUARTPortRead] = HEADER_CMD;
      }
      else if (CmdState[CurrentUARTPortRead] == HEADER_CMD)
      {   
          Cmd[CurrentUARTPortRead] = c;        
          CmdCheckSum[CurrentUARTPortRead] ^= c;
          CmdState[CurrentUARTPortRead] = HEADER_DATA;
      }
      else if (CmdState[CurrentUARTPortRead] == HEADER_DATA && InBufferOffset[CurrentUARTPortRead] < CmdDataSize[CurrentUARTPortRead])
      {             
          CmdCheckSum[CurrentUARTPortRead] ^= c;
          InBuffer[InBufferOffset[CurrentUARTPortRead]++][CurrentUARTPortRead] = c;
      }
      else if (CmdState[CurrentUARTPortRead] == HEADER_DATA && InBufferOffset[CurrentUARTPortRead] >= CmdDataSize[CurrentUARTPortRead])
      {   
         if (CmdCheckSum[CurrentUARTPortRead] == c) 
         {
           ConversionBytesIntoStructure();
         }
                          
         CmdState[CurrentUARTPortRead] = IDLE;
         cc = 0;
      }
    }
  }
}

#ifdef UART_HARDWARE_DEBUG
  void WriteDebug(uint8_t c)
  {
    UARTSerialize(UART_DEBUG_PORT,c);
  }
  
  void WriteDebug(const uint8_t* buffer, size_t size)
  {
    while (size--) 
    {
      WriteDebug(*buffer++);
    }
  }
  
  inline void WriteDebug(unsigned long n) 
  { 
    WriteDebug((uint8_t)n); 
  }
  
  inline void WriteDebug(long n) 
  { 
    WriteDebug((uint8_t)n); 
  }
  
  inline void WriteDebug(unsigned int n) 
  { 
    WriteDebug((uint8_t)n); 
  }
  
  inline void WriteDebug(int n) 
  { 
    WriteDebug((uint8_t)n); 
  }
  
  void WriteDebug(const char *buffer, size_t size) 
  {
    WriteDebug((const uint8_t *)buffer, size);
  }
  
  void WriteDebug(const char *str) 
  {
        if (str != NULL) 
        {
          WriteDebug((const uint8_t *)str, strlen(str));
        }
  }
  
  void PrintDebugNumber(unsigned long n, uint8_t base) 
  {
    char buf[8 * sizeof(long) + 1]; 
    char *str = &buf[sizeof(buf) - 1];
  
    *str = '\0';
  
    if (base < 2) 
    {
      base = 10;
    }
  
    do 
    {
      unsigned long m = n;
      n /= base;
      char c = m - base * n;
      *--str = c < 10 ? c + '0' : c + 'A' - 10;
    } 
    while(n);
  
    WriteDebug(str);
  }
  
  void PrintDebugFloat(double number, uint8_t digits) 
  { 
    if (isnan(number)) 
    {
      PrintDebug("nan");
      return;
    }
    if (isinf(number)) 
    {
      PrintDebug("inf");
      return;
    }
    if (number > 4294967040.0) 
    {
      PrintDebug ("ovf"); 
      return;
    } 
    if (number <-4294967040.0) 
    {
      PrintDebug ("ovf");  
      return;
    }
    
    if (number < 0.0)
    {
       PrintDebug('-');
       number = -number;
    }
  
    double rounding = 0.5;
    for (uint8_t i=0; i<digits; ++i)
    {
      rounding /= 10.0;
    }
    
    number += rounding;
  
    unsigned long int_part = (unsigned long)number;
    double remainder = number - (double)int_part;
    PrintDebug(int_part);
  
    if (digits > 0) 
    {
      PrintDebug("."); 
    }
  
    while (digits-- > 0)
    {
      remainder *= 10.0;
      int toPrint = int(remainder);
      PrintDebug(toPrint);
      remainder -= toPrint; 
    } 
  }
  
  void PrintDebug(const char str[])
  {
    WriteDebug(str);
  }
  
  void PrintDebug(char c)
  {
    WriteDebug(c);
  }
  
  void PrintDebug(unsigned char b, uint8_t base)
  {
    PrintDebug((unsigned long) b, base);
  }
  
  void PrintDebug(int n, uint8_t base)
  {
    PrintDebug((long) n, base);
  }
  
  void PrintDebug(unsigned int n, uint8_t base)
  {
    PrintDebug((unsigned long) n, base);
  }
  
  void PrintDebug(long n, uint8_t base)
  {
    if (base == 0) 
    {
      WriteDebug(n);
    } 
    else if (base == 10) 
    {
      if (n < 0) 
      {
        PrintDebug('-');
        n = -n;
        PrintDebugNumber(n, 10);
      }
      PrintDebugNumber(n, 10);
    } 
    else 
    {
      PrintDebugNumber(n, base);
    }
  }
  
  void PrintDebug(unsigned long n, uint8_t base)
  {
    if (base == 0) 
    {
      WriteDebug(n);
    }
    else 
    {
      PrintDebugNumber(n, base);
    }
  }
  
  void PrintDebug(double n, uint8_t digits)
  {
    PrintDebugFloat(n, digits);
  }
  
  void PrintDebug(const __FlashStringHelper *ifsh)
  {
    PGM_P p = reinterpret_cast<PGM_P>(ifsh);
    while (1) 
    {
      unsigned char c = pgm_read_byte(p++);
      if (c == 0) 
      {
        break;
      }
      WriteDebug(c);
    }
  }
  
  void PrintDebug(const String &s)
  {
    for (uint16_t i = 0; i < s.length(); i++) 
    {
      WriteDebug(s[i]);
    }
  }
  
  void PrintDebug(unsigned char n)
  {
    PrintDebug(n, DEC);
  }
  
  void PrintDebug(int n)
  {
    PrintDebug(n, DEC);
  }
  
  void PrintDebug(unsigned int n)
  {
    PrintDebug(n, DEC);
  }
  
  void PrintDebug(long n)
  {
    PrintDebug(n, DEC);
  }
  
  void PrintDebug(unsigned long n)
  {
    PrintDebug(n, DEC);
  }

  void PrintDebug(double n)
  {
    PrintDebug(n, 2);
  }
#endif
//********************** /Объявление публичных функции  *********************//

